# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from functools import wraps


def provide_method_name(func, provide_instance=False):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if provide_instance:
            kwargs['instance'] = args[0]
        kwargs['method'] = func.name if hasattr(func, 'name') else func.__name__
        return func(*args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from .helpers import make_list

from six import string_types


class State(object):
    """
    Atomic class for transition class.
    """
    def __init__(self, name, id=None, value=None):
        """
        :param name: name of the state
        :param id: integer or string basing on field
        :param value: display name of the state
        :type name: string
        :type id: integer or string
        :type value: string
        """
        self.name = name
        self.id = id
        self.value = value

    def __eq__(self, other):
        if isinstance(other, string_types):
            return self.name == other
        elif isinstance(other, State):
            return self.name == other.name

    def __str__(self):
        return self.name
    __unicode__ = __str__

    def __repr__(self):
        return "<{0}: '{1}'>".format(self.__class__.__name__, self.name)


class Transition(object):
    """
    Atomic class for handler, holds data about transition.
    """
    def __init__(self, trigger, start_state, end_state, before=None, after=None):
        """
        :param trigger: function that triggers instance state change
        :param start_state: name of a start state
        :param end_state: name of a end state
        :param before: methods that should be fired before trigger function
        :param after: methods that should be fired after trigger function
        :type trigger: string or handler method
        :type start_state: string
        :type end_state: string
        :type before: single method, function or list of function, methods
        :type after: single method, function or list of function, methods
        """
        self.trigger = trigger
        self.start_state = start_state
        self.end_state = end_state
        self.before = make_list(before) if before else []
        self.after = make_list(after) if after else []

    def __str__(self):
        return "From '{0}' -> '{1}', trigger: {2}".format(self.start_state or '*', self.end_state, self.trigger)
    __unicode__ = __str__

    def __repr__(self):
        return "<{0}: From '{1}' -> '{2}'>".format(self.__class__.__name__, self.start_state or '*', self.end_state)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from .data_classes import State, Transition
from .decorators import provide_method_name
from .exceptions import TransitionNotPossibleError, WrongTriggerTypeError
from .state_map import StateMachineMap

from collections import defaultdict
from six import string_types


class StateCache(object):
    """
    StateCache class works as a State factory and cache, at the given
    handler_key which is generated automatically at the init of a handler
    and by state_name you retrieve the state you want.

    The `type` parameter indicates the type of a field, if there are no
    kwargs provided for state initialization then StateCache inserts
    values basing on `type`.
    If the type is `int` then it will try to insert item with id + 1,
    If the type is `str` then it will try to insert item with state_name key

    Example:
    defaultdict(dict, {
        'TestHandler_state': {
            'accepted': <State: 'accepted'>,
            'notaccepted': <State: 'notaccepted'>,
            'rejected': <State: 'rejected'>
        }
    })
    """
    _state_objects = defaultdict(dict)

    @classmethod
    def get_state(cls, handler_key, state_name, field_type=None, **kwargs):
        """
        Create or get a state basing on the `handler_key` and `state_name`.
        If kwargs are not provided then generate auto values.
        """
        if state_name in cls._state_objects[handler_key].keys():
            return cls._state_objects[handler_key][state_name]
        if not kwargs and field_type:
            state = cls._create_auto_values(handler_key, state_name, field_type)
        else:
            state = State(state_name, **kwargs)
        cls._state_objects[handler_key][state_name] = state
        return state

    @classmethod
    def _create_auto_values(cls, handler_key, state_name, field_type):
        """
        This method automatically creates dummy state values if only
        transitions are provided for handler.
        """
        if field_type == 'int':
            states = cls._state_objects[handler_key].values()
            ids = sorted(state.id for state in states) or [1]
            id = ids[-1]
            value = state_name.lower().title()
            return State(state_name, id=id, value=value)
        elif field_type == 'str':
            value = state_name.lower().title()
            return State(state_name, id=state_name, value=value)


class TransitionProcessMixin(object):
    def process(self, transition, **kwargs):
        """
        Transition processing starts with checking if transition can be made:
        1) if no - exception `TransitionNotPossibleError` is raised.
        2) if yes - then:
            - kwargs parameter will be passed to every method in steps below.
            - every function in transition `before` is called.
            - handler `trigger` is called.
            - handler updates instance state value.
            - every function in transition `after` is called.
        """
        if not self.can_make_transition(transition):
            raise TransitionNotPossibleError("Can't make transition from '{0}' to '{1}', instance value: '{2}'".format(
                transition.start_state, transition.end_state, self.get_instance_field_value()
            ))
        self._call_before(transition, **kwargs)
        self._call_trigger(transition, **kwargs)
        self.update_state(transition)
        self._call_after(transition, **kwargs)

    def _call_before(self, transition, **kwargs):
        """
        Call every function in `before` attribute. Kwargs are
        provided from `process` method.
        """
        for before_function in transition.before:
            before_function(**kwargs)

    def _call_trigger(self, transition, **kwargs):
        """
        Call trigger function from handler class. Because of the
        nature of implementation of handler classes, the trigger
        function in the runtime is prefixed with '_'. Kwargs are
        provided from `process` method.
        """
        trigger_name = transition.trigger.__name__
        handler_trigger = getattr(self, '_' + trigger_name)
        handler_trigger(**kwargs)

    def _call_after(self, transition, **kwargs):
        """
        Call every function in `after` attribute. Kwargs are
        provided from `process` method.
        """
        for after_function in transition.after:
            after_function(**kwargs)

    def can_make_transition(self, transition):
        """
        Check if instance qualifies for transition. It checks instance value
        and compares it with transition start state.

        :return: True or False
        """
        if transition.start_state == '*':
            pass

        if self.get_instance_field_value() == transition.start_state.id:
            return True
        return False

    def update_state(self, transition):
        setattr(self.instance, self.field_name, transition.end_state.id)


class TransitionAddingMixin(object):
    def add_transitions(self):
        """
        After subclassing `BaseStateHandler` override this method to add transitions for example it could look
        like this:
            def add_transitions(self):
                self.add_transition('accept', 'initial', 'accepted')
                self.add_transition('reject', 'initial', 'rejected')
        """

    def add_transition(self, trigger, start_state, end_state, before=None, after=None):
        """
        General method for adding transitions to your handler, first it checks if the name
        exists in already defined methods, if yes, throw an error, if not, proceed to
        creating states and transition, map transition to `self.states_map` then create and
        assign a template function for this trigger.
        Template function will take the place of the current trigger function, when the one implemented
        on the handler will be moved to the name with underscore.
        """
        handler_method = self._check_trigger(trigger)
        start_state = StateCache.get_state(self.handler_key, start_state, field_type=self.field_type)
        end_state = StateCache.get_state(self.handler_key, end_state, field_type=self.field_type)
        transition = Transition(handler_method, start_state, end_state, before, after)
        self.states_map.add_transition(transition)

        def add_trigger(self, method, **kwargs):
            transition = self.states_map.transitions[method]
            kwargs['instance'] = self.instance
            self.process(transition, **kwargs)
        add_trigger.__name__ = handler_method.__name__

        setattr(self.__class__, "_" + trigger, handler_method)
        setattr(self.__class__, trigger, provide_method_name(add_trigger))

    def _check_trigger(self, trigger):
        """
        Checks whether the provided trigger is a method on a handler, if it exists and is a callable.
        trigger: can be a callable or a string that refers to a method implemented on handler
        """
        if isinstance(trigger, string_types):
            if hasattr(self.__class__, trigger) and callable(getattr(self.__class__, trigger)):
                return getattr(self.__class__, trigger)
        elif callable(getattr(self.__class__, trigger.__name__)):
            return trigger
        raise WrongTriggerTypeError("Provided {0} trigger is not a string or a callable".format(trigger))


class BaseStateHandler(TransitionProcessMixin, TransitionAddingMixin):
    def __init__(self, field_name, field_type, **kwargs):
        self.field_name = field_name
        self.field_type = field_type
        self.handler_key = "{0}_{1}".format(self.__class__, field_name)
        self.states_map = StateMachineMap()
        self.state_choices = kwargs.get('state_choices', [])
        self.instance = None
        if self.state_choices:
            self._init_state_choices()
        self.add_transitions()

    def _init_state_choices(self):
        for state_values_row in self.state_choices:
            id, value, name = state_values_row
            state = StateCache.get_state(self.handler_key, name, id=id, value=value)
            self.states_map.add_state(state)

    def get_instance_field_value(self):
        """Get value of a instance field value"""
        return getattr(self.instance, self.field_name)

    @property
    def choices(self):
        choices = []
        for node in self.states_map.machine_map.values():
            choices.append((node.state.id, node.state.value))
        return choices



from setuptools import setup

try:
    long_description = open('README.rst').read()
except IOError:
    long_description = ''

setup(
    name='django-state-machines',
    version='0.1.2',
    description='Django finite state machine implementation',
    author='Karol Sztajerwald',
    author_email='sztajerwaldkarol@gmail.com',
    url='https://github.com/Backscratcher/django-state-machines',
    keywords="django",
    packages=['django_state_machines', ],
    include_package_data=True,
    zip_safe=False,
    license='MIT License',
    platforms=['any'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        "Framework :: Django",
        "Framework :: Django :: 1.8",
        "Framework :: Django :: 1.9",
        "Framework :: Django :: 1.10",
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.db import models

from .decorators import provide_method_name

import copy


class StateMachineMixin(object):
    def __init__(self, handler, *args, **kwargs):
        self.state_choices = kwargs.pop('state_choices', None)

        if self.state_choices:
            choices = []
            for id, value, state_helper_name in self.state_choices:
                choices.append((id, value))
            kwargs['choices'] = choices

        super(StateMachineMixin, self).__init__(*args, **kwargs)
        self.handler_class = handler

    def deconstruct(self):
        name, path, args, kwargs = super(StateMachineMixin, self).deconstruct()
        kwargs['handler'] = self.handler_class
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, **kwargs):
        super(StateMachineMixin, self).contribute_to_class(cls, name, **kwargs)
        field_type = self._get_field_type(cls)
        self.handler = self.handler_class(name, field_type, state_choices=self.state_choices)
        if not self.choices:
            self.choices = self.handler.choices

        def property_handler(self, method, instance):
            handler_name = '_{0}'.format(method)
            if handler_name not in self.__dict__:
                field_name, rest = method.split('_')
                field = self._meta.get_field(field_name)
                handler_copy = copy.deepcopy(field.handler)
                handler_copy.instance = instance
                self.__dict__[handler_name] = handler_copy
            return self.__dict__[handler_name]
        property_handler.name = "{0}_handler".format(self.handler.field_name)

        setattr(cls, "{0}_handler".format(self.handler.field_name), property(provide_method_name(property_handler, provide_instance=True)))

    def _get_field_type(self, cls):
        if isinstance(cls, models.IntegerField):
            return 'int'
        elif isinstance(cls, models.CharField):
            return 'str'


class StateMachineIntegerField(StateMachineMixin, models.IntegerField):
    pass


class StateMachineCharField(StateMachineMixin, models.CharField):
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals


class WrongStateError(Exception):
    pass


class DuplicateTransitionTriggerError(Exception):
    pass


class NoSuchStateError(Exception):
    pass


class TransitionNotPossibleError(Exception):
    pass


class WrongTriggerTypeError(Exception):
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from .data_classes import State
from .exceptions import DuplicateTransitionTriggerError, NoSuchStateError

from six import string_types


class StateMapNode(object):
    """
    Atomic node class for `StateMachineMapDict` it holds informations about single state and
    where you can proceed: to which states and with which transitions and from which states and
    with which transitions you can get to this state.
    """
    def __init__(self, state):
        self.state = state
        self.allowed_transitions = []
        self.allowed_states = []
        self.from_transitions = []
        self.from_states = []

    def __repr__(self):
        return "<Node {0}: allowed_states={1}, allowed_transitions={2}, from_states={3}, from_transitions={4}>".format(
            self.state, self.allowed_states, self.allowed_transitions, self.from_states, self.from_transitions)


class StateMachineMapDict(dict):
    """ Dictionary that holds objects of `StateMapNode`."""
    def alter_key(self, key):
        if isinstance(key, State):
            return key.name
        return key

    def __getitem__(self, key):
        return super(StateMachineMapDict, self).__getitem__(self.alter_key(key))

    def __contains__(self, key):
        return super(StateMachineMapDict, self).__contains__(self.alter_key(key))


class StateMachineMap(object):
    """
    Instances of this class have two attributes:
     - Dictionary transitions, which serves as a fast hash-map for getting transitions.
     - Dictionary machine_map , which is a representation of a `StateMachineMapDict`.
    Purpose of this class is to hold whole structure of state machine and to give
    nice API to for example: get allowed states for some state value.
    """
    def __init__(self):
        self.transitions = {}
        self.machine_map = StateMachineMapDict()

    def _add_states(self, states):
        for state in states:
            self.machine_map[state.name] = StateMapNode(state)

    def _update_states(self, transition):
        transition_start = transition.start_state
        self.machine_map[transition_start.name].allowed_transitions.append(transition)
        self.machine_map[transition_start.name].allowed_states.append(transition.end_state)
        transition_end = transition.end_state
        self.machine_map[transition_end.name].from_transitions.append(transition)
        self.machine_map[transition_end.name].from_states.append(transition.start_state)

    def _map_transition(self, transition):
        states = []
        if transition.start_state not in self.machine_map.keys():
            states.append(transition.start_state)
        if transition.end_state not in self.machine_map.keys():
            states.append(transition.end_state)
        self._add_states(states)
        self._update_states(transition)

    def add_transition(self, transition):
        callback_name = transition.trigger if \
            isinstance(transition.trigger, string_types) else transition.trigger.__name__
        if callback_name in self.transitions.keys():
            raise DuplicateTransitionTriggerError(
                "Trigger with name '{0}' already exists.".format(callback_name))

        self.transitions[callback_name] = transition
        self._map_transition(transition)

    def add_state(self, state):
        if state not in self.machine_map.keys():
            self._add_states([state])

    def get_state_info(self, state):
        """
        Method returns `StateMapNode` of the provided state from `machine_map`.
        """
        try:
            return self.machine_map[state]
        except KeyError:
            raise NoSuchStateError("'{0}' state not found in map.".format(state))

    def get_states_info(self):
        return self.machine_map.values()
